"""Radiance Daylight Coefficient Grid-Based Analysis Recipe."""
from ..recipeutil import write_extra_files
from ..recipedcutil import (write_rad_files_daylight_coeff,
    get_commands_w_groups_daylight_coeff, skymtx_to_gendaymtx)
from ..recipedcutil import get_commands_w_groups_daylight_coeff
from .._gridbasedbase import GenericGridBased
from ..parameters import get_radiance_parameters_grid_based
from ...sky.skymatrix import SkyMatrix
from ....futil import write_to_file
from ...analysisgrid import AnalysisGrid
from ...parameters.rfluxmtx import RfluxmtxParameters
from ....hbsurface import HBSurface
from ...command.rfluxmtx import Rfluxmtx
from ...command.rcontrib import Rcontrib
from ...command.dctimestep import Dctimestep
from ...command.oconv import Oconv
from ...command.rmtxop import Rmtxop, RmtxopMatrix
from ..recipeutil import glz_srf_to_window_group

import os
from collections import namedtuple

def sky_receiver(filepath, density, ground_file_format=None, sky_file_format=None):
    """Create a receiver sky for daylight coefficient studies."""
    if not (ground_file_format and sky_file_format):
        return Rfluxmtx.default_sky_ground(filepath, sky_type='r{}'.format(density))
    else:
        return Rfluxmtx.default_sky_ground(
            filepath, sky_type='r{}'.format(density),
            ground_file_format=ground_file_format,
            sky_file_format=sky_file_format)

def coeff_matrix_commands(output_name, receiver, rad_files, sender, points_file=None,
                          number_of_points=None, rfluxmtx_parameters=None):
    """Returns radiance commands to create coefficient matrix.

    Args:
        output_name: Output file name.
        receiver: A radiance file to indicate the receiver. In view matrix it will be the
        window group and in daylight matrix it will be the sky.
        rad_files: A collection of Radiance files that should be included in the scene.
        sender: A collection of files for senders if senders are radiance geometries
            such as window groups (Default: '-').
        points_file: Path to point file which will be used instead of sender.
        number_of_points: Number of points in points_file as an integer.
        rfluxmtx_parameters: Radiance parameters for Rfluxmtx command using a
            RfluxmtxParameters instance (Default: None).
    """
    sender = sender or '-'
    rad_files = rad_files or ()
    number_of_points = number_of_points or 0
    rfluxmtx = Rfluxmtx()

    if sender == '-':
        assert points_file, \
            ValueError('You have to set the points_file when sender is not defined.')

    # -------------- set the parameters ----------------- #
    rfluxmtx.rfluxmtx_parameters = rfluxmtx_parameters

    # -------------- set up the sender objects ---------- #
    # '-' in case of view matrix, window group in case of
    # daylight matrix. This is normally the receiver file
    # in view matrix
    rfluxmtx.sender = sender

    # points file are the senders in view matrix
    rfluxmtx.number_of_points = number_of_points
    rfluxmtx.points_file = points_file

    # --------------- set up the  receiver --------------- #
    # This will be the window for view matrix and the sky for
    # daylight matrix. It makes sense to make a method for each
    # of thme as they are pretty repetitive
    # Klems full basis sampling
    rfluxmtx.receiver_file = receiver

    # ------------- add radiance geometry files ----------------
    # For view matrix it's usually the room itself and the materials
    # in case of each view analysis rest of the windows should be
    # blacked! In case of daylight matrix it will be the context
    # outside the window.
    rfluxmtx.rad_files = rad_files

    # output file address\name
    rfluxmtx.output_matrix = output_name

    return rfluxmtx

def _get_commands_imageless_glare(
        project_name, sky_density, project_folder, window_group, skyfiles, inputfiles,
        points_file, total_point_count, blkmaterial, wgsblacked, rfluxmtx_parameters, 
        occupancy, DGP, sGA, window_group_count=0, window_groupfiles=None, 
        reuse_daylight_mtx=False, counter=None):
    """Get commands for the daylight coefficient recipe.

    This function is used by get_commands_scene_daylight_coeff and
    get_commands_w_groups_daylight_coeff. You usually don't want to use this function
    directly.
    """
    commands = []
    result_files = []
    # unpack inputs
    opqfiles, glzfiles, wgsfiles, extrafiles = inputfiles
    # edited to only have one sky file unlike original dc grid based recipe
    sky_mtx_total = skyfiles

    for scount, state in enumerate(window_group.states):
        # 2.3.Generate daylight coefficients using rfluxmtx
        # collect list of radiance files in the scene for both total and direct
        if counter:
            p = ((counter[0] + scount - 1.0) / counter[1]) * 100
            c = int(p / 10)
            commands.append(
                ':: Done with {} of {} ^|{}{}^| ({:.2f}%%)'.format(
                    counter[0] + scount - 1, counter[1], '#' * c,
                    '-' * (10 - c), float(p)
                )
            )
        commands.append('::')
        commands.append(
            ':: start of the calculation for {}, {}. State {} of {}'.format(
                window_group.name, state.name, scount + 1, window_group.state_count
            )
        )
        commands.append('::')

        if scount != 0 or not window_groupfiles:
            # in case window group is not already provided
            window_groupfiles = (wgsfiles[window_group_count].fp[scount],)

        rflux_scene = (
            f for fl in
            (window_groupfiles, opqfiles.fp, extrafiles.fp,
             blkmaterial, wgsblacked)
            for f in fl)

        rflux_scene_blacked = (
            f for fl in
            (window_groupfiles, opqfiles.fpblk, extrafiles.fpblk,
             blkmaterial, wgsblacked)
            for f in fl)

        d_matrix = 'result/matrix/normal_{}..{}..{}.dc'.format(
            project_name, window_group.name, state.name)

        d_matrix_direct = 'result/matrix/black_{}..{}..{}.dc'.format(
            project_name, window_group.name, state.name)

        if not os.path.isfile(os.path.join(project_folder, d_matrix)) \
                or not os.path.isfile(os.path.join(project_folder, d_matrix_direct)) \
                or not reuse_daylight_mtx:
            rad_files = tuple(os.path.relpath(f, project_folder) for f in rflux_scene)
            sender = '-'
            receiver = sky_receiver(
                os.path.join(project_folder, 'sky/rfluxSky.rad'), sky_density
            )

            commands.append(':: :: 1. calculating daylight matrices')
            commands.append('::')

            commands.append(':: :: [1/2] scene daylight matrix')
            commands.append(
                ':: :: rfluxmtx - [sky] [points] [wgroup] [blacked wgroups] [scene]'
                ' ^> [dc.mtx]'
            )
            commands.append('::')

            # sampling_rays_count = 1 based on @sariths studies
            rflux = coeff_matrix_commands(
                d_matrix, os.path.relpath(receiver, project_folder), rad_files, sender,
                os.path.relpath(points_file, project_folder), total_point_count,
                rfluxmtx_parameters
            )
            commands.append(rflux.to_rad_string())

            rad_files_blacked = tuple(os.path.relpath(f, project_folder)
                                        for f in rflux_scene_blacked)

            commands.append(':: :: [2/2] black scene daylight matrix')
            commands.append(
                ':: :: rfluxmtx - [sky] [points] [wgroup] [blacked wgroups] '
                '[blacked scene] ^> [black dc.mtx]'
            )
            commands.append('::')

            original_value = int(rfluxmtx_parameters.ambient_bounces)
            rfluxmtx_parameters.ambient_bounces = 1
            rflux_direct = coeff_matrix_commands(
                d_matrix_direct, os.path.relpath(receiver, project_folder),
                rad_files_blacked, sender,
                os.path.relpath(points_file, project_folder),
                total_point_count, rfluxmtx_parameters
            )
            commands.append(rflux_direct.to_rad_string())
            rfluxmtx_parameters.ambient_bounces = original_value

            commands.append('::')
        else:
            commands.append(':: :: 1. reusing daylight matrices')
            commands.append('::')

        commands.append(':: :: 2. calculating DGP')
        commands.append(':: :: dcglare [views] [dcdirect] [dctotal] [sky_mtx] ^> [views.dgp]')
        commands.append('::')
        dcglare_dgp = 'views..{}..{}.dgp'.format(window_group.name, state.name)
        dcglare = 'c:\\radiance\\bin\\dcglare -vf {} {} {} {} > {}'.format(
            os.path.relpath(points_file, project_folder), 
            d_matrix_direct, 
            d_matrix, 
            sky_mtx_total,
            dcglare_dgp)
        commands.append(dcglare)
        commands.append('::')

        commands.append(':: :: 2. calculating glare autonomy')
        commands.append(':: :: dcglare [views] [occupancy] [dcdirect] [dctotal] [sky_mtx] ^> [views.ga]')
        commands.append('::')
        dcglare_occ_dgp = 'occupied..{}..{}.ga'.format(window_group.name, state.name)
        dcglare = 'c:\\radiance\\bin\\dcglare -vf {} -sf "{}" -l {} {} {} {} > {}'.format(
            os.path.relpath(points_file, project_folder), 
            occupancy, 
            DGP, 
            d_matrix_direct, 
            d_matrix, 
            sky_mtx_total, 
            dcglare_occ_dgp)
        commands.append(dcglare)
        commands.append('::')

        commands.append(':: :: 3. calculating spatial glare autonomy')
        commands.append('::')
        spatial_ga = 'spatial..{}..{}.ga'.format(window_group.name, state.name)
        spatial_ga_command = ' | '.join([
            'c:\\radiance\\bin\\rmtxop {}'.format(dcglare_occ_dgp),
            'getinfo -',
            'rcalc -e "$1=floor($1+{})"'.format(sGA),
            'total -m > {}'.format(spatial_ga)])
        commands.append(spatial_ga_command)
        commands.append('::')

        result_files.append(
                os.path.join(project_folder, dcglare_dgp)
            )
        result_files.append(
                os.path.join(project_folder, dcglare_occ_dgp)
            )
        result_files.append(
                os.path.join(project_folder, spatial_ga)
            )

        commands.append(
            ':: end of calculation for {}, {}'.format(window_group.name, state.name))
        commands.append('::')

    return commands, result_files

def get_commands_scene_imageless_glare(
        project_name, sky_density, project_folder, skyfiles, inputfiles,
        points_file, total_point_count, rfluxmtx_parameters, occupancy, DGP, sGA, 
        reuse_daylight_mtx=False, total_count=1):
    """Get commands for the static windows in the scene.

    Use get_commands_w_groups_daylight_coeff to get the commands for the rest of the
    scene.

    Args:
        project_name: A string to generate uniqe file names for this project.
        sky_density: Sky density for this study.
        project_folder: Path to project_folder.
        skyfiles: Collection of path to sky files. The order must be (sky_mtx_total,
            sky_mtx_direct, analemma, sunlist, analemmaMtx). You can use get_commands_sky
            function to generate this list.
        inputfiles: Input files for this study. The order must be (opqfiles, glzfiles,
            wgsfiles, extrafiles). Each files object is a namedtuple which includes
            filepath to radiance files under fp and filepath to backed out files under
            fpblk.
        points_file: Path to points_file.
        total_point_count: Number of total points inside points_file.
        rfluxmtx_parameters: An instance of rfluxmtx_parameters for daylight matrix.
        reuse_daylight_mtx: A boolean not to include the commands for daylight matrix
            calculation if they already exist inside the folder.
    """
    # unpack inputs
    opqfiles, glzfiles, wgsfiles, extrafiles = inputfiles

    if len(wgsfiles) > 0:
        # material is the first file
        blkmaterial = [wgsfiles[0].fpblk[0]]
        # collect files for blacked geometries for all window groups
        wgsblacked = [f.fpblk[1] for c, f in enumerate(wgsfiles)]
    else:
        # there is no window group, return an empty tuple
        blkmaterial = ()
        wgsblacked = ()

    window_group = glz_srf_to_window_group()
    window_groupfiles = glzfiles.fp

    commands, results = _get_commands_imageless_glare(
        project_name, sky_density, project_folder, window_group, skyfiles,
        inputfiles, points_file, total_point_count, blkmaterial, wgsblacked,
        rfluxmtx_parameters, occupancy, DGP, sGA, 0, window_groupfiles, 
        reuse_daylight_mtx, (1, total_count))

    return commands, results

def get_commands_sky(project_folder, sky_matrix, reuse=True):
    """Get list of commands to generate the skies.

    1. total sky matrix

    This method generates sun matrix under project_folder/sky and return the command
    to generate skies number 1 and 2.

    Returns a namedtuple for (output_files, commands)
    output_files in a namedtuple itself (sky_mtx_total, sky_mtx_direct, analemma,
        sunlist, analemmaMtx).
    """

    SkyCommands = namedtuple('SkyCommands', 'output_files commands')

    commands = []

    # # 2.1.Create sky matrix.
    sky_matrix.mode = 0
    sky_mtx_total = 'sky/{}.smx'.format(sky_matrix.name)

    # add command for total sky matrix.
    if hasattr(sky_matrix, 'isSkyMatrix'):
        gdm = skymtx_to_gendaymtx(sky_matrix, project_folder)
        if gdm:
            note = ':: {} sky matrix'.format('total')
            commands.extend((note, gdm))
        sky_matrix.mode = 0
    else:
        # sky vector
        raise TypeError('You must use a SkyMatrix to generate the sky.')

    of = sky_mtx_total

    return SkyCommands(commands, of)

class ImagelessGlareGridBased(GenericGridBased):
    """Grid based daylight coefficient analysis recipe.

    Attributes:
        sky_mtx: A radiance SkyMatrix or SkyVector. For an SkyMatrix the analysis
            will be ran for the analysis period.
        analysis_grids: A list of Honeybee analysis grids. Daylight metrics will
            be calculated for each analysisGrid separately.
        simulation_type: 0: Illuminance(lux), 1: Radiation (kWh), 2: Luminance (Candela)
            (Default: 0)
        radiance_parameters: Radiance parameters for this analysis. Parameters
            should be an instance of RfluxmtxParameters.
        hb_objects: An optional list of Honeybee surfaces or zones (Default: None).
        sub_folder: Analysis subfolder for this recipe. (Default: "gridbased_imagelessglare").
    """

    def __init__(self, sky_mtx, analysis_grids, occupancy, DGP=0.4, sGA=0.05,
                 radiance_parameters=None, reuse_daylight_mtx=True, hb_objects=None,
                 sub_folder="gridbased_imagelessglare"):
        """Create an annual recipe."""
        GenericGridBased.__init__(
            self, analysis_grids, hb_objects, sub_folder
        )

        self.sky_matrix = sky_mtx
        self.occupancy = occupancy
        self.DGP = DGP
        self.sGA = sGA
        self.radiance_parameters = radiance_parameters
        self.radiance_parameters.irradiance_calc = True
        self.reuse_daylight_mtx = reuse_daylight_mtx

    @property
    def sky_matrix(self):
        """Get and set sky definition."""
        return self._sky_matrix

    @sky_matrix.setter
    def sky_matrix(self, new_sky):
        assert hasattr(new_sky, 'isRadianceSky'), \
            '%s is not a valid Honeybee sky.' % type(new_sky)
        assert not new_sky.is_point_in_time, \
            TypeError('Sky for daylight coefficient recipe must be a sky matrix.')
        self._sky_matrix = new_sky.duplicate()

    @property
    def radiance_parameters(self):
        """Radiance parameters for annual analysis."""
        return self._radiance_parameters

    @radiance_parameters.setter
    def radiance_parameters(self, par):
        if not par:
            # set RfluxmtxParameters as default radiance parameter for annual analysis
            par = get_radiance_parameters_grid_based(0, 1).dmtx

        assert hasattr(par, 'isRfluxmtxParameters'), \
            TypeError('Expected RfluxmtxParameters not {}'.format(type(par)))

        self._radiance_parameters = par

    @property
    def sky_density(self):
        """Radiance sky type e.g. r1, r2, r4."""
        return "r{}".format(self.sky_matrix.sky_density)

    @property
    def total_runs_count(self):
        """Number of total runs for all window groups and states."""
        return sum(wg.state_count for wg in self.window_groups) + 1  # 1 for base case

    def preproc_commands(self):
        """Add echo in front of comments in batch file comments."""
        cmd = [c for c in self._commands if c]
        cmd = ['echo ' + c if c[:2] == '::' else c for c in cmd]
        return ['@echo off'] + cmd

    def _add_commands(self, skycommands, commands):
        """Check if the commands should be added to self._commands."""
        if self.reuse_daylight_mtx:
            if not skycommands:
                for f in self._result_files:
                    if not os.path.isfile(f):
                        self._commands.extend(commands)
                        break
            else:
                # there are changes in the sky.
                # matrices multiplication needs to be recalculated.
                self._commands.extend(commands)
        else:
            # there are changes in the sky.
            # matrices multiplication needs to be recalculated.
            self._commands.extend(commands)

    def write(self, target_folder, project_name='untitled', header=True,
              transpose=False):
        """Write analysis files to target folder.

        Args:
            target_folder: Path to parent folder. Files will be created under
                target_folder/gridbased. use self.sub_folder to change subfolder name.
            project_name: Name of this project as a string.
            header: A boolean to include the header lines in commands.bat. header
                includes PATH and cd toFolder
        Returns:
            Full path to command.bat
        """
        # 0.prepare target folder
        # create main folder target_folder/project_name
        self._commands = []
        self._result_files = []

        project_folder = \
            super(GenericGridBased, self).write_content(
                target_folder, project_name, False, subfolders=['tmp', 'result/matrix']
            )

        # write geometry and material files
        opqfiles, glzfiles, wgsfiles = write_rad_files_daylight_coeff(
            project_folder + '/scene', project_name, self.opaque_rad_file,
            self.glazing_rad_file, self.window_groups_rad_files
        )
        # additional radiance files added to the recipe as scene
        extrafiles = write_extra_files(self.scene, project_folder + '/scene', True)

        # 0.write points
        points_file = self.write_analysis_grids(project_folder, project_name)

        # 2.write batch file
        if header:
            self._commands.append(self.header(project_folder))

        # # 2.1.Create sky matrix.
        # # 2.2. Create sun matrix
        skycommands, skyfiles = get_commands_sky(project_folder, self.sky_matrix,
                                                 reuse=True)

        self._commands.extend(skycommands)

        # for each window group - calculate total, direct and direct-analemma results
        # calculate the contribution of glazing if any with all window groups blacked
        inputfiles = opqfiles, glzfiles, wgsfiles, extrafiles
        commands, results = get_commands_scene_imageless_glare(
            project_name, self.sky_matrix.sky_density, project_folder, skyfiles,
            inputfiles, points_file, self.total_point_count, self.radiance_parameters,
            self.occupancy, self.DGP, self.sGA, self.reuse_daylight_mtx, 
            self.total_runs_count)

        self._result_files.extend(
            os.path.join(project_folder, str(result)) for result in results
        )

        self._add_commands(skycommands, commands)

        if self.window_groups:
            # calculate the contribution for all window groups
            commands, results = get_commands_w_groups_daylight_coeff(
                project_name, self.sky_matrix.sky_density, project_folder,
                self.window_groups, skyfiles, inputfiles, points_file,
                self.total_point_count, self.radiance_parameters,
                self.reuse_daylight_mtx, self.total_runs_count, transpose=transpose)

            self._add_commands(skycommands, commands)
            self._result_files.extend(
                os.path.join(project_folder, str(result)) for result in results
            )

        # # 2.5 write batch file
        batch_file = os.path.join(project_folder, 'commands.bat')

        # add echo to commands and write them to file
        write_to_file(batch_file, '\n'.join(self.preproc_commands()))

        return batch_file

    def results(self):
        return self._result_files
